import git
import os
import shutil
import primary_language


# parses through commits in reverse chronological order, hence the flipping of the terms for the arguments
def commit_analysis(repo, cutoff_date, start_date):
    print("Analyzing Commits...")
    commits_info = []
    for commit in repo.iter_commits():
        # if too far back, break
        if commit.committed_datetime > start_date:
            continue
        if commit.committed_datetime < cutoff_date:
            break
        commit_info = {
            "commit_hash": commit.hexsha,
            "author_name": commit.author.name,
            "author_email": commit.author.email,
            "authored_date": commit.authored_datetime,
            "committer_name": commit.committer.name,
            "committer_email": commit.committer.email,
            "commit_date": commit.committed_datetime,
            "message": commit.message,
            "is_merge": len(commit.parents) > 1,
        }
        # author/committer org information 
        commit_info['author_org'] = commit_info["author_email"].split("@")[-1].split(".")[0]
        commit_info['committer_org'] = commit_info["committer_email"].split("@")[-1].split(".")[0]
        # some more effort to get this information
        commit_info["branches"] = repo.git.branch(
            "--contains", commit_info["commit_hash"]
        )
        # diff information
        diffs = commit.diff(
            commit.parents[0] if commit.parents else git.NULL_TREE, create_patch=True
        )
        commit_info["diff_info"] = diff_analysis(diffs)
        # print(commit_info)
        commits_info.append(commit_info)
    return commits_info


def diff_analysis(diffs):
    diff_objects = []
    for diff in diffs:
        diff_info = {
            "lines_added": sum(
                1
                for line in diff.diff.decode("utf-8").split("\n")
                if line.startswith("+") and not line.startswith("+++")
            ),
            "lines_deleted": sum(
                1
                for line in diff.diff.decode("utf-8").split("\n")
                if line.startswith("-") and not line.startswith("---")
            ),
            "parent_filepath": diff.a_path,
            "child_filepath": diff.b_path,
            "change_type": diff.change_type,
            "new_file": diff.new_file,
            "deleted_file": diff.deleted_file,
            "renamed_file": diff.renamed,
            #'diff': diff.diff.decode('utf-8')
        }
        diff_objects.append(diff_info)
    return diff_objects



# __init__.py

"""
    repo_lifecycles is a small, lightweight data collection tool to support the longitudinal
    analysis of software repositories. 
"""

from .main import repo_lifecycle

__all__ = [
    "repo_lifecycle"
]



import os
import shutil

import git

def temp_clone(vcs_link, temp_location):
    """
    ARGS
    vcs_link : url link to upstream repo vcs
    temp_location : filepath to where the repo should be cloned to 

    RETURNS
    repo : the GitRepository object of the cloned repo 
    repo_path : the filepath to the cloned repository
    """
    os.makedirs(temp_location)
    repo_path = temp_location
    repo = git.Repo.clone_from(vcs_link, repo_path)
    print(f"Successfully Cloned {vcs_link}")
    return repo, repo_path


def delete_clone(temp_location):
    """
    ARGS
    temp_location : filepath to the cloned repository 

    RETURNS
    whether or not the deletion was a success
    """
    if os.path.exists(temp_location):
        shutil.rmtree(temp_location)
        print(f"{temp_location} has been deleted.")
        return 0
    else:
        print("No clone at location")
        return 1



import os
import json
from collections import defaultdict


def language_sizes(repo_path):
    json_file_path = os.path.join(
        os.path.dirname(__file__), "resources", "file_extensions.json"
    )
    with open(json_file_path, "r") as json_file:
        language_extensions = json.load(json_file)
    language_sizes = defaultdict(int)
    for root, _, files in os.walk(repo_path):
        for file in files:
            ext = file.split(".")[-1]
            if ext in language_extensions.keys():
                file_path = os.path.join(root, file)
                with open(file_path, "r", errors="ignore") as f:
                    lines = f.readlines()
                    language_sizes[language_extensions[ext]] += len(lines)
    return language_sizes



import argparse
import sys
import os
import datetime
import subprocess

import pandas as pd

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "lib")))
from lib import cloning, commit_analysis
from lib import primary_language as pl

def repo_lifecycle(
    vcs_link, clone_location, from_date, to_date, to_save=False, to_csv=True
):
    """
    ARGS
        vcs_link (str): Version control system link of the repository.
        from_date (datetime): The start date for commit analysis.
        to_date (datetime): The end date for commit analysis.
        to_save (boolean): whether to save the project repository at the to_date; default FALSE
        to_csv (boolean): whether to save the commit information in a csv; default TRUE
    """
    try:
        # download the repository
        temp_repo, temp_repo_path = cloning.temp_clone(vcs_link, clone_location)
        project_name = vcs_link.split("/")[-1]
        project_owner = vcs_link.split("/")[-2]

        # collect data on prior commits
        commits_array = commit_analysis.commit_analysis(temp_repo, from_date, to_date)
        commits_df = pd.DataFrame.from_records(commits_array)
        if to_csv:
            commits_df.to_csv(
                f"{project_owner}_{project_name}_{from_date.strftime('%Y-%m-%d')}_to_{to_date.strftime('%Y-%m-%d')}.csv",
                index=False,
            )

        if to_save:
            _commit = commits_df.iloc[0].to_dict()
            temp_repo_path = checkout_version(
                project_owner, project_name, _commit, temp_repo_path
            )
            print(temp_repo_path)

            # program/code analysis of the project at a specific point in time
            '''
             language_breakdown = pl.language_sizes(
                temp_repo_path
            )  # breakdown is in terms of LoC
            print(language_breakdown)
            '''
            # TODO: here goes the linter/analysis implementation
    except Exception as e:
        print(f"An error occurred: {e}")
    # clean up
    finally:
        if not to_save:
            cloning.delete_clone(temp_repo_path)


def checkout_version(project_owner, project_name, commit, repo_path):
    """
    ARGS
    project_owner : the entity that owns the git repo 
    project_name : the name of the git repo 
    commit : the commit object to checkout from 
    repo_path : the filepath of the cloned repo 

    RETURNS
    new_repo_path : the updated filepath of the checked-out repo 
    """
    # update the repo with the correct checkout
    subprocess.run(
        ["git", "checkout", commit["commit_hash"]], cwd=repo_path, check=True
    )
    new_repo_path = (
        "/".join(repo_path.split("/")[:-1])
        + f"{commit['commit_date'].strftime('%Y-%m-%d')}_{project_owner}_{project_name}_{commit['commit_hash']}"
    )
    os.rename(repo_path, new_repo_path)
    return new_repo_path


if __name__ == "__main__":
    # setting the defaults for the test run
    LOCATION = "tmp/new/"
    parser = argparse.ArgumentParser(description="repository search")
    parser.add_argument("repolink", help="The repository hosting")
    args = parser.parse_args()
    cst = datetime.timezone(datetime.timedelta(hours=-6))
    from_date = datetime.datetime(2024, 10, 10, 1, 52, 32, tzinfo=cst)
    to_date = datetime.datetime(2024, 11, 20, 1, 52, 32, tzinfo=cst)
    # getting the information for the search
    repo_lifecycle(args.repolink, LOCATION, from_date, to_date, to_save=True)
